'use strict';

var core = require('@react-monaco/core');
var react = require('react');
var vscodeOniguruma = require('vscode-oniguruma');
var vscodeTextmate = require('vscode-textmate');

const wasmUrl = 'vscode-oniguruma/release/onig.wasm';
const [setupTextmate, tmConfig] = core.makeConfigurable({
    onigurumaWasmUrl: core.repoUrlOf(core.monacoConfig('repo'), wasmUrl),
    baseUrl: core.monacoAssetsOf('tm/'),
    languages: ';yaml;yaml-embedded;yaml-1.3;yaml-1.2;yaml-1.1;yaml-1.0;xsl;xml;asp-vb-net;typescriptreact;typescript;jsdoc.ts.injection;jsdoc.js.injection;sql;swift;shell-unix-bash;shaderlab;searchresult;scss;sassdoc;ruby;rust;cshtml;pug;rst;magicregexp;magicpython;r;powershell;php;html;md-math;md-math-inline;md-math-fence;md-math-block;make;log;objective-c;objective-c++;less;lua;perl6;perl;tex;markdown-latex-combined;latex;cpp-grammar-bailout;bibtex;markdown;html;html-derivative;ini;java;javascriptreact;javascript;handlebars;hlsl;snippets;jsonl;jsonc;json;fsharp;ignore;git-rebase;git-commit;go;diff;docker;groovy;julia;dart;csharp;css;batchfile;coffeescript;clojure;platform;cuda-cpp;cpp;cpp.embedded.macro;c;dbml;kotlin;'
});
const isTmSupportLanguage = (id)=>tmConfig('languages').includes(`;${id};`);

// import * as monacoNsps from 'monaco-editor'
// as described in issue: https://github.com/NeekSandhu/monaco-textmate/issues/5
const TMToMonacoToken = (editor, scopes)=>{
    let scopeName = '';
    if (!scopes[0]) return '';
    // get the scope name. Example: cpp , java, haskell
    for(let i = scopes[0].length - 1; i >= 0; i -= 1){
        const char = scopes[0][i];
        if (char === '.') {
            break;
        }
        scopeName = char + scopeName;
    }
    // iterate through all scopes from last to first
    for(let i = scopes.length - 1; i >= 0; i -= 1){
        const scope = scopes[i];
        if (scope == null) continue;
        /**
     * Try all possible tokens from high specific token to low specific token
     *
     * Example:
     * 0 meta.function.definition.parameters.cpp
     * 1 meta.function.definition.parameters
     *
     * 2 meta.function.definition.cpp
     * 3 meta.function.definition
     *
     * 4 meta.function.cpp
     * 5 meta.function
     *
     * 6 meta.cpp
     * 7 meta
     */ for(let i = scope.length - 1; i >= 0; i -= 1){
            const char = scope[i];
            if (char === '.') {
                const token = scope.slice(0, i);
                if (// @ts-ignore
                // biome-ignore lint/complexity/useLiteralKeys: <explanation>
                editor['_themeService']._theme._tokenTheme._match(`${token}.${scopeName}`)._foreground > 1) {
                    return `${token}.${scopeName}`;
                }
                if (// @ts-ignore
                // biome-ignore lint/complexity/useLiteralKeys: <explanation>
                editor['_themeService']._theme._tokenTheme._match(token)._foreground > 1) {
                    return token;
                }
            }
        }
    }
    return '';
};

class TokenizerState {
    _ruleStack;
    constructor(_ruleStack){
        this._ruleStack = _ruleStack;
    }
    get ruleStack() {
        return this._ruleStack;
    }
    clone() {
        return new TokenizerState(this._ruleStack);
    }
    equals(other) {
        if (!other || !(other instanceof TokenizerState) || other !== this || other._ruleStack !== this._ruleStack) {
            return false;
        }
        return true;
    }
}
/**
 * Wires up monaco-editor with monaco-textmate
 *
 * @param _monaco monaco namespace this operation should apply to (usually the `monaco` global unless you have some other setup)
 * @param registry TmGrammar `Registry` this wiring should rely on to provide the grammars
 * @param languages `Map` of language ids (string) to TM names (string)
 */ function wireTmGrammars(_monaco, registry, languages, editor) {
    return Promise.all(Array.from(languages.keys()).map(async (languageId)=>{
        const scopeName = languages.get(languageId);
        if (scopeName == null) return;
        const grammar = await registry.loadGrammar(scopeName);
        if (grammar == null) return;
        _monaco.languages.setTokensProvider(languageId, {
            getInitialState: ()=>new TokenizerState(vscodeTextmate.INITIAL),
            // @ts-ignore
            tokenize: (line, state)=>{
                // @ts-ignore
                const res = grammar.tokenizeLine(line, state.ruleStack);
                return {
                    endState: new TokenizerState(res.ruleStack),
                    tokens: res.tokens.map((token)=>({
                            ...token,
                            // TODO: At the moment, monaco-editor doesn't seem to accept array of scopes
                            scopes: editor ? TMToMonacoToken(editor, token.scopes) : token.scopes[token.scopes.length - 1]
                        }))
                };
            }
        });
    }));
}

class TextmateEventsDelegator extends core.BaseEventsDelegator {
    scopeName = [
        'Textmate',
        'color: orange'
    ];
    wasmKey = 'wasm/onigasm';
    wireGrammars = {};
    providers = {};
    code;
    vsRegistry;
    vsOnigurumaLib;
    constructor(props){
        super(props);
        this.register('prepareAssets').register('onAssetVSCodeWasm', `asset:${this.wasmKey}`).register('mounting').register('prepareModel').register('createModel').register('mountEditor');
    }
    get baseUrl() {
        return this.options.baseUrl || tmConfig('baseUrl');
    }
    get onigurumaWasmUrl() {
        return this.options.onigurumaWasmUrl || tmConfig('onigurumaWasmUrl');
    }
    get wasmAsset() {
        return {
            key: this.wasmKey,
            url: new URL(this.onigurumaWasmUrl, this.baseUrl),
            priority: 100,
            type: 'wasm'
        };
    }
    prepareAssets = ({ preloadAssets })=>{
        preloadAssets.push(this.wasmAsset);
        this.debug(`add asset '${this.wasmKey}'`);
    };
    onAssetVSCodeWasm = async ({ task, handle })=>{
        if (task.chunks != null) {
            try {
                await vscodeOniguruma.loadWASM(task.chunks.buffer);
                this.vsOnigurumaLib = Promise.resolve({
                    createOnigScanner: (patterns)=>new vscodeOniguruma.OnigScanner(patterns),
                    createOnigString: (s)=>new vscodeOniguruma.OnigString(s)
                });
                this.debug('vscode-oniguruma wasm loaded');
            } catch (err) {
                this.debug('load vscode-oniguruma wasm error', err);
            }
        }
        if (this.vsOnigurumaLib == null) {
            this.debug('vscode-oniguruma lib instance is null');
        } else {
            this.vsRegistry = this.newVsCodeRegistry(this.vsOnigurumaLib);
            this.debug('create grammar registry');
        }
        handle();
    };
    mounting = ()=>{
        if (typeof this.options.mounting === 'function') {
            this.options.mounting();
        }
    };
    prepareModel = (params)=>{
        const { provider: providerCallback } = this.options;
        const { language, extname } = params;
        let languageId;
        let provider;
        if (providerCallback != null) {
            provider = providerCallback(params);
        }
        if (provider != null) {
            languageId = provider.languageId;
            this.providers[languageId] = provider;
        } else {
            if (language != null) {
                languageId = language.id;
            } else if (extname != null) {
                const ext = (extname || '').replace(/^./gm, '').toLowerCase();
                if (isTmSupportLanguage(ext)) {
                    languageId = ext;
                }
            }
        }
        this.code = this.createCodeSet(languageId, params, provider);
        this.debug('create code set', this.code);
    };
    // 基于 prepareModel 派生，用于优先创建对应的 Code
    createCodeSet = (languageId, params, provider)=>{
        const { onChange, filter } = this.options;
        const timestamp = Date.now();
        const { monaco, language } = params;
        let isActive = false;
        const extname = (params.extname || '').toLowerCase();
        if (!languageId) {
            onChange?.({
                isActive,
                languageId: 'plaintext',
                extname,
                timestamp
            });
            return undefined;
        }
        isActive = true;
        // 不存在的 language ，优先朝 monaco 注册语言
        if (language == null) {
            monaco.languages.register({
                id: languageId,
                extensions: [
                    extname ?? ''
                ]
            });
        }
        onChange?.({
            isActive,
            languageId,
            extname,
            timestamp
        });
        const code = {
            ...this.selectTextmateScope(languageId, extname, language, provider),
            isWired: false,
            language,
            extname,
            languageId,
            provider: this.providers[languageId]
        };
        if (filter) {
            return filter(code);
        }
        return code;
    };
    /**
   * 基于 vscode 新规范，形成正规的 scopeName 和 tmName
   * @param languageId
   * @param extname
   * @param language
   * @param provider
   */ selectTextmateScope = (languageId, extname, language, provider)=>{
        let tmName = languageId;
        let scopeName = `source${extname}`;
        switch(extname){
            case '.tsx':
                tmName = 'typescriptreact';
                break;
            case '.jsx':
                tmName = 'javascriptreact';
                break;
        }
        switch(language?.id){
            case 'python':
                tmName = 'magicpython';
                scopeName = 'source.python';
                break;
            case 'scss':
                scopeName = 'source.css.scss';
                break;
            case 'rust':
                scopeName = 'source.rust';
                break;
        }
        if (provider?.scopeName) {
            scopeName = provider.scopeName;
        }
        return {
            scopeName,
            tmName
        };
    };
    createModel = ({ monaco, editor })=>this.wireVsCodeState(monaco, editor);
    mountEditor = ({ monaco, editor })=>this.wireVsCodeState(monaco, editor);
    wireVsCodeState = (_monaco, editor)=>{
        if (_monaco == null || editor == null || this.code == null || this.vsRegistry == null) {
            return;
        }
        const { isWired, languageId, scopeName, tmName, language } = this.code;
        if (isWired) return;
        if (isTmSupportLanguage(tmName) || this.providers[languageId]) {
            this.debug('wire grammar', {
                languageId,
                scopeName,
                tmName
            });
            if (!this.wireGrammars[tmName]) {
                this.wireGrammars[tmName] = true;
                const grammars = new Map();
                grammars.set(languageId, scopeName);
                // @ts-ignore typeof monacoNsps
                wireTmGrammars(_monaco, this.vsRegistry, grammars, editor).then(()=>{
                    this.debug('wire grammar done');
                });
            }
        }
    };
    newVsCodeRegistry = (onigLib)=>new vscodeTextmate.Registry({
            onigLib,
            loadGrammar: async (scopeName)=>{
                if (this.code == null) return null;
                try {
                    const baseUrl = this.baseUrl;
                    let url = new URL(`${this.code.tmName}.tmLanguage.json`, baseUrl);
                    if (this.code.provider != null) {
                        url = new URL(this.code.provider.url, baseUrl);
                    }
                    this.debug(`load '${scopeName}' grammar from ${url}`);
                    const resp = await fetch(url, {
                        cache: 'force-cache'
                    });
                    const text = await resp.text();
                    this.debug(`load '${scopeName}' grammar success, size:`, text.length);
                    // vscode-textmate 新版本，基于文件名来判断是  json 合适 plist
                    return vscodeTextmate.parseRawGrammar(text, url.toString());
                } catch (err) {
                    console.log(`load ${scopeName} grammar error`, err);
                }
                return null;
            }
        });
}

const TextmateInjection = (props)=>{
    const { emitterRef, setError, extendThemes, lifecycleId } = core.useMonaco();
    const delegatorRef = react.useRef(null);
    // biome-ignore lint/correctness/useExhaustiveDependencies: <explanation>
    react.useLayoutEffect(()=>{
        try {
            delegatorRef.current = new TextmateEventsDelegator({
                ...props,
                mounting: extendThemes
            });
            delegatorRef.current.inject(emitterRef.current);
        } catch (err) {
            setError(err);
        }
        return ()=>{
            delegatorRef.current?.eject(emitterRef.current);
            delegatorRef.current = null;
        };
    }, [
        lifecycleId
    ]);
    return null;
};

exports.TextmateEventsDelegator = TextmateEventsDelegator;
exports.TextmateInjection = TextmateInjection;
exports.isTmSupportLanguage = isTmSupportLanguage;
exports.setupTextmate = setupTextmate;
exports.tmConfig = tmConfig;
